#  _______  ______    _______  _______  _______ 
# |       ||    _ |  |       ||       ||       |
# |    ___||   | ||  |   _   ||   _   ||    ___|
# |   |___ |   |_||_ |  | |  ||  | |  ||   | __ 
# |    ___||    __  ||  |_|  ||  |_|  ||   ||  |
# |   |    |   |  | ||       ||       ||   |_| |
# |___|    |___|  |_||_______||_______||_______|

import os
import numpy as np
from inspect import signature
from typing import Tuple, List, Union, Optional, Any, TypeVar, cast
from froog.gpu import get_device, upload_tensor, download_tensor, is_buffer

T = TypeVar('T', bound='Tensor')

class Tensor:
    did_float_warning = False
    ops = {}
    ops_gpu = {}

    def __init__(self, data: Union[List, np.ndarray, Any], gpu: bool = False):
        if isinstance(data, list): data = np.array(data, dtype=np.float32)
        elif is_buffer(data): self.gpu = True
        elif not isinstance(data, np.ndarray): raise TypeError(f"Error constructing tensor with {data}")
        if isinstance(data, np.ndarray):
            if data.dtype != np.float32 and not Tensor.did_float_warning:
                if os.getenv("WARNING") == "1": print(f"warning, {data.shape} isn't float32. float64 needed for numerical jacobian")
                if not os.getenv("DEBUG") == "1": Tensor.did_float_warning = True
            self.gpu = False
        self.data = data
        self.grad: Optional[Tensor] = None
        self._ctx = None
        if gpu: self.gpu_()

    def __repr__(self) -> str: return f"Tensor data: {self.data}, gradients: {self.grad.data if self.grad else None}"
    def assign(self, x: T) -> None: self.data = x.data

    @property
    def shape(self) -> Tuple[int, ...]:
        if self.gpu:
            device = get_device()
            if device is not None and hasattr(device, 'buffer_metadata'):
                buffer_id = id(self.data)
                if buffer_id in device.buffer_metadata: return device.buffer_metadata[buffer_id]['shape']
            try:
                data = download_tensor(self)
                return data.shape
            except Exception as e:
                print(f"Warning: Failed to get shape from GPU tensor: {e}")
                return (1,)
        return self.data.shape

    @property
    def size(self, dim=None) -> Union[int, Tuple[int, ...]]:
        if dim is not None: return self.shape[dim]
        return int(np.prod(self.shape))

    @property
    def ndim(self) -> int: return len(self.shape)

    @property
    def transpose(self) -> T:
        if isinstance(self.data, np.ndarray): return Tensor(self.data.T, gpu=self.gpu)
        else:
            cpu_tensor = self.to_cpu()
            return Tensor(cpu_tensor.data.T, gpu=self.gpu)

    @property
    def dtype(self) -> np.dtype:
        if self.gpu:
            device = get_device()
            if device is not None and hasattr(device, 'buffer_metadata'):
                buffer_id = id(self.data)
                if buffer_id in device.buffer_metadata: return device.buffer_metadata[buffer_id]['dtype']
            return np.float32
        return self.data.dtype

    @property
    def is_gpu(self) -> bool: return self.gpu

    @staticmethod
    def zeros(*shape: int) -> T: return Tensor(np.zeros(shape, dtype=np.float32))

    @staticmethod
    def ones(*shape: int) -> T: return Tensor(np.ones(shape, dtype=np.float32))

    @staticmethod
    def randn(*shape: int) -> T: return Tensor(np.random.randn(*shape).astype(np.float32))

    @staticmethod
    def eye(dim: int) -> T: return Tensor(np.eye(dim).astype(np.float32))

    @staticmethod
    def arange(start: Union[int, float], stop: Optional[Union[int, float]] = None, step: Union[int, float] = 1) -> T:
        if stop is None:
            stop = start
            start = 0
        return Tensor(np.arange(start, stop, step, dtype=np.float32))

    def flatten(self) -> T: return Tensor(self.data.reshape(-1), gpu=self.gpu)
    def detach(self) -> T: return Tensor(self.data.copy(), gpu=self.gpu)
    def view(self, *shape: int) -> T: return Tensor(self.data.reshape(shape), gpu=self.gpu)
    def to_float(self) -> T: return Tensor(self.data.astype(np.float32), gpu=self.gpu)
    def to_int(self) -> T: return Tensor(self.data.astype(np.int32), gpu=self.gpu)
    def to_bool(self) -> T: return Tensor(self.data.astype(bool), gpu=self.gpu)

    def unsqueeze(self, dim: int) -> T:
        shape = list(self.shape)
        if dim < 0: dim = len(shape) + 1 + dim
        shape.insert(dim, 1)
        return Tensor(self.data.reshape(shape), gpu=self.gpu)

    def squeeze(self, dim: Optional[int] = None) -> T:
        if dim is None: return Tensor(self.data.squeeze(), gpu=self.gpu)
        shape = list(self.shape)
        if dim < 0: dim = len(shape) + dim
        if 0 <= dim < len(shape) and shape[dim] == 1: shape.pop(dim)
        return Tensor(self.data.reshape(shape), gpu=self.gpu)

    def backward(self, allow_fill: bool = True) -> None:
        if self._ctx is None: return
        if self.grad is None and allow_fill:
            assert self.shape == (1,)
            self.grad = Tensor(np.ones(self.shape, dtype=self.dtype), gpu=self.gpu)
        assert self.grad is not None
        grads = self._ctx.backward(self._ctx, self.grad.data)
        if len(self._ctx.parents) == 1: grads = [grads]
        for t, g in zip(self._ctx.parents, grads):
            if g is None: continue
            t_shape = t.shape
            if is_buffer(g):
                device = get_device()
                if device is not None and hasattr(device, 'buffer_metadata'):
                    buffer_id = id(g)
                    if buffer_id in device.buffer_metadata: g_shape = device.buffer_metadata[buffer_id]['shape']
                    else:
                        try:
                            g_cpu = download_tensor(g)
                            g_shape = g_cpu.shape
                        except:
                            print(f"Warning: Could not determine shape of gradient in {self._ctx}")
                            g_shape = t_shape
            else: g_shape = g.shape
            if g_shape != t_shape:
                print(f"grad shape must match tensor shape in {self._ctx}, {g_shape} != {t_shape}")
                assert False
            t.grad = Tensor(g)
            t.backward(allow_fill=False)

    def mean(self) -> T:
        div = Tensor(np.array([1 / self.size], dtype=np.float32), gpu=self.gpu)
        return self.sum().mul(div)

    def sqrt(self) -> T:
        root = Tensor(np.zeros(self.shape, dtype=np.float32) + 0.5, gpu=self.gpu)
        return self.pow(root)

    def div(self, y: T) -> T:
        root = Tensor(np.zeros(self.shape, dtype=np.float32) - 1, gpu=self.gpu)
        return self.mul(y.pow(root))
    
    def to_cpu(self) -> T:
        if not self.gpu: return cast(T, self)
        data = download_tensor(self)
        ret = Tensor(data)
        if self.grad: ret.grad = self.grad.to_cpu()
        return ret
    
    def gpu_(self) -> None:
        if not self.gpu and (device := get_device()) is not None and device.name != "CPU":
            self.data = upload_tensor(self.data)
            self.gpu = True
            if self.grad: self.grad.gpu_()

    def to_gpu(self) -> T:
        if (device := get_device()) is None or device.name == "CPU": raise Exception("no gpu support! install pyopencl or use a Metal-compatible device")
        if self.gpu: return cast(T, self)
        gpu_data = upload_tensor(self.data)
        ret = Tensor(gpu_data)
        ret.gpu = True
        if self.grad: ret.grad = self.grad.to_gpu()
        return ret

class Function:
    def __init__(self, *tensors: Tensor) -> None:
        self.parents = tensors
        self.saved_tensors: List[Any] = []

    def save_for_backward(self, *x: Any) -> None:
        self.saved_tensors.extend(x)

    def apply(self, *x: Any, **kwargs: Any) -> Tensor:
        op = self
        ctx = op(*x)
        params = signature(op.forward).parameters
        for p in params.values():
            if p.default is not p.empty: setattr(ctx, p.name, p.default)
        for k, v in kwargs.items(): setattr(ctx, k, v)
        ret = Tensor(op.forward(ctx, *[t.data for t in x], **kwargs))
        ret._ctx = ctx
        return ret

def register(name: str, fxn: Any, gpu: bool = False) -> None:
    if gpu:
        setattr(Tensor, name, lambda self, *x, **kwargs: fxn.apply(fxn, self, *x, **kwargs))
        Tensor.ops_gpu[name] = fxn
    else: Tensor.ops[name] = fxn

    def dispatch(self: Tensor, *x: Any, **kwargs: Any) -> Tensor:
        try:
            op_func = (Tensor.ops_gpu if self.gpu else Tensor.ops)[name]
            return op_func.apply(op_func, self, *x, **kwargs)
        except Exception as e:
            print(f"Error in {name} operation: {e}")
            if os.getenv("DEBUG") == "1":
                print(f"  Self: {self}")
                for i, arg in enumerate(x): print(f"  Arg {i}: {arg}")
                print(f"  Kwargs: {kwargs}")
            raise

    setattr(Tensor, name, dispatch)

    if name in ['add', 'sub', 'mul', 'div']:
        setattr(Tensor, "__%s__" % name, dispatch)
        setattr(Tensor, "__i%s__" % name, lambda self, x: self.assign(dispatch(self, x)))

if (device := get_device()) is not None and device.name != "CPU":
    if device.__class__.__name__ == "MetalDevice":
        try: import froog.gpu.metal.ops_metal
        except ImportError: 
            if os.getenv("DEBUG") == "1": print("Failed to import Metal operations")
    elif device.__class__.__name__ == "OpenCLDevice":
        try: import froog.gpu.cl.ops_cl
        except ImportError: 
            if os.getenv("DEBUG") == "1": print("Failed to import OpenCL operations")


#  _______  ______    _______  _______  _______ 
# |       ||    _ |  |       ||       ||       |
# |    ___||   | ||  |   _   ||   _   ||    ___|
# |   |___ |   |_||_ |  | |  ||  | |  ||   | __ 
# |    ___||    __  ||  |_|  ||  |_|  ||   ||  |
# |   |    |   |  | ||       ||       ||   |_| |
# |___|    |___|  |_||_______||_______||_______|

import numpy as np
from functools import lru_cache
import pathlib, hashlib, os, tempfile, urllib
from typing import Tuple

def fetch(url: str) -> pathlib.Path:
  if url.startswith(("/", ".")): return pathlib.Path(url)
  else: fp = pathlib.Path("_cache_dir") / "froog" / "downloads" / (hashlib.md5(url.encode('utf-8')).hexdigest())
  if not fp.is_file():
    with urllib.request.urlopen(url, timeout=10) as r:
      assert r.status == 200
      total_length = int(r.headers.get('content-length', 0))
      (path := fp.parent).mkdir(parents=True, exist_ok=True)
      with tempfile.NamedTemporaryFile(dir=path, delete=False) as f:
        while chunk := r.read(16384): f.write(chunk)
        f.close()
        if (file_size:=os.stat(f.name).st_size) < total_length: raise RuntimeError(f"fetch size incomplete, {file_size} < {total_length}")
        pathlib.Path(f.name).rename(fp)
  return fp

def fetch_mnist() -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
  import gzip
  parse = lambda file: np.frombuffer(gzip.open(file).read(), dtype=np.uint8).copy()
  BASE_URL = "https://storage.googleapis.com/cvdf-datasets/mnist/"
  X_train = parse(fetch(f"{BASE_URL}train-images-idx3-ubyte.gz"))[0x10:].reshape((-1, 28*28)).astype(np.float32)
  Y_train = parse(fetch(f"{BASE_URL}train-labels-idx1-ubyte.gz"))[8:].astype(np.int8)
  X_test = parse(fetch(f"{BASE_URL}t10k-images-idx3-ubyte.gz"))[0x10:].reshape((-1, 28*28)).astype(np.float32)
  Y_test = parse(fetch(f"{BASE_URL}t10k-labels-idx1-ubyte.gz"))[8:].astype(np.int8)
  return X_train, Y_train, X_test, Y_test

def mask_like(like: np.ndarray, mask_inx: np.ndarray, mask_value: float = 1.0) -> np.ndarray:
  mask = np.zeros_like(like).reshape(-1) # flatten
  mask[mask_inx] = mask_value            # fill 
  return mask.reshape(like.shape)

@lru_cache
def get_im2col_index(oy: int, ox: int, cin: int, H: int, W: int) -> np.ndarray:
  idx_channel = np.tile(np.arange(cin).repeat(H*W), oy*ox)
  idx_y = np.tile(np.arange(H).repeat(W), oy*ox*cin) + np.arange(oy).repeat(ox*cin*H*W)
  idx_x = np.tile(np.arange(W), oy*ox*cin*H) + np.tile(np.arange(ox), oy).repeat(cin*H*W)
  OY, OX = oy+(H-1), ox+(W-1)
  idx = idx_channel * OY * OX + idx_y * OX + idx_x
  return idx

@lru_cache
def rearrange_col2im_index(oy: int, ox: int, cin: int, H: int, W: int) -> np.ndarray:
  idx = get_im2col_index(oy, ox, cin, H, W)
  r_idx = np.zeros((np.max(idx)+1, H*W), dtype=idx.dtype)-1
  for i,x in enumerate(idx):
    for j in range(H*W):
      if r_idx[x,j] == -1:
        r_idx[x,j] = i
        break
  return r_idx

# im2col convolution helpers
def im2col(x: np.ndarray, H: int, W: int) -> np.ndarray:
  bs, cin, oy, ox = x.shape[0], x.shape[1], x.shape[2]-(H-1), x.shape[3]-(W-1)
  idx = get_im2col_index(oy, ox, cin, H, W)
  tx = x.reshape(bs, -1)[:, idx]

  # all the time is spent here
  # np.ravel() flattens the array into a 1-dimensional shape
  tx = tx.ravel()
  return tx.reshape(-1, cin*W*H)

def col2im(tx: np.ndarray, H: int, W: int, OY: int, OX: int) -> np.ndarray:
  oy, ox = OY-(H-1), OX-(W-1)
  bs = tx.shape[0] // (oy * ox)
  channels_in = tx.shape[1] // (H * W)

  ridx = rearrange_col2im_index(oy, ox, channels_in, H, W)
  # -1 has to be 0s
  x = np.pad(tx.reshape(bs, -1), ((0,0),(0,1)))[:, ridx].sum(axis=2)
  return x.reshape(bs, channels_in, OY, OX)


#  _______  ______    _______  _______  _______ 
# |       ||    _ |  |       ||       ||       |
# |    ___||   | ||  |   _   ||   _   ||    ___|
# |   |___ |   |_||_ |  | |  ||  | |  ||   | __ 
# |    ___||    __  ||  |_|  ||  |_|  ||   ||  |
# |   |    |   |  | ||       ||       ||   |_| |
# |___|    |___|  |_||_______||_______||_______|

import numpy as np
from typing import List
from froog.tensor import Tensor

class Optimizer:
  def __init__(self, params: List[Tensor]) -> None:
    self.params = params

class SGD(Optimizer):
  """
  Stochastic Gradient Descent
  """
  def __init__(self, params: List[Tensor], lr: float = 0.001, weight_decay: float = 0, clip_value: float = 0) -> None:
    super(SGD, self).__init__(params)
    self.lr = Tensor([lr], gpu=params[0].gpu if params else False)
    self.weight_decay = weight_decay
    self.clip_value = clip_value

  def step(self) -> None:
    for t in self.params:
      if t.grad is None:
        continue
      
      if t.gpu:
        from froog.gpu import get_device, download_tensor, upload_tensor
        
        # device = get_device()
        t_cpu = download_tensor(t)
        grad_cpu = download_tensor(t.grad)
        lr_cpu = download_tensor(self.lr)
        
        if self.weight_decay > 0:
          grad_cpu += self.weight_decay * t_cpu
        
        if self.clip_value > 0:
          grad_cpu = np.clip(grad_cpu, -self.clip_value, self.clip_value)
        
        t_cpu -= grad_cpu * lr_cpu
        t.data = upload_tensor(t_cpu)
      else:
        if self.weight_decay > 0:
          t.grad.data += self.weight_decay * t.data
        
        if self.clip_value > 0:
          t.grad.data = np.clip(t.grad.data, -self.clip_value, self.clip_value)
        
        t -= t.grad * self.lr

class Adam(Optimizer):  
  """
  Default ADAM optimizer from https://arxiv.org/pdf/1412.6980.pdf algorithm
  """
  def __init__(self, params: List[Tensor], lr: float = 0.001, b1: float = 0.9, b2: float = 0.999, eps: float = 1e-8, max_grad: float = 10.0) -> None:
    super(Adam, self).__init__(params)
    self.lr = lr
    self.b1 = b1
    self.b2 = b2
    self.eps = eps
    self.t = 0
    self.max_grad = max_grad
    self.on_gpu = any(t.gpu for t in self.params if t is not None)
    
    if self.on_gpu:
      from froog.gpu import download_tensor
      self.m = [np.zeros_like(download_tensor(t.data)) for t in self.params]
      self.v = [np.zeros_like(download_tensor(t.data)) for t in self.params]
    else:
      self.m = [np.zeros_like(t.data) for t in self.params]
      self.v = [np.zeros_like(t.data) for t in self.params]

  def step(self) -> None:
    from froog.gpu import download_tensor, upload_tensor
    
    self.t += 1
    a = self.lr * (np.sqrt(1 - np.power(self.b2, self.t)) / (1 - np.power(self.b1, self.t)))
      
    for i, t in enumerate(self.params):
      if t.grad is None:
        continue
        
      if t.gpu:
        try:
          t_data_cpu = download_tensor(t.data)
          grad_cpu = download_tensor(t.grad.data)
          
          if np.isnan(grad_cpu).any() or np.isinf(grad_cpu).any():
            print(f"Warning: NaN or Inf detected in gradients for parameter {i}")
            grad_cpu = np.nan_to_num(grad_cpu, nan=0.0, posinf=self.max_grad, neginf=-self.max_grad)
          
          if self.max_grad > 0:
            grad_cpu = np.clip(grad_cpu, -self.max_grad, self.max_grad)
          
          self.m[i] = self.b1 * self.m[i] + (1 - self.b1) * grad_cpu
          self.v[i] = self.b2 * self.v[i] + (1 - self.b2) * np.square(grad_cpu)
          
          denom = np.sqrt(self.v[i]) + self.eps
          update = a * self.m[i] / denom
          
          if np.isnan(update).any() or np.isinf(update).any():
            print(f"Warning: NaN or Inf detected in update for parameter {i}")
            max_update = np.finfo(np.float32).max / 100
            update = np.nan_to_num(update, nan=0.0, posinf=max_update, neginf=-max_update)
          
          t_data_cpu -= update
          
          if np.isnan(t_data_cpu).any() or np.isinf(t_data_cpu).any():
            print(f"Warning: NaN or Inf detected in parameter {i} after update")
            max_val = np.finfo(np.float32).max / 10
            t_data_cpu = np.nan_to_num(t_data_cpu, nan=0.0, posinf=max_val, neginf=-max_val)
          
          t.data = upload_tensor(t_data_cpu)
        except Exception as e:
          print(f"Error in Adam update for GPU tensor {i}: {e}")
          continue
      else:
        if self.max_grad > 0:
          np.clip(t.grad.data, -self.max_grad, self.max_grad, out=t.grad.data)
          
        self.m[i] = self.b1 * self.m[i] + (1 - self.b1) * t.grad.data
        self.v[i] = self.b2 * self.v[i] + (1 - self.b2) * np.square(t.grad.data)
        t.data -= a * self.m[i] / (np.sqrt(self.v[i]) + self.eps)

class RMSprop(Optimizer):
  """
  RMSprop optimizer with epsilon for numerical stability.
  """
  def __init__(self, params: List[Tensor], decay: float = 0.9, lr: float = 0.001, eps: float = 1e-8) -> None:
    super(RMSprop, self).__init__(params)
    self.lr = lr
    self.decay = decay
    self.eps = eps
    self.v: List[np.ndarray] = [np.zeros_like(t.data) for t in self.params]

  def step(self) -> None:
    for i, t in enumerate(self.params):
      self.v[i] = self.decay * self.v[i] + (1 - self.decay) * np.square(t.grad.data)
      t.data -= self.lr / (np.sqrt(self.v[i]) + self.eps) * t.grad.data


#  _______  ______    _______  _______  _______ 
# |       ||    _ |  |       ||       ||       |
# |    ___||   | ||  |   _   ||   _   ||    ___|
# |   |___ |   |_||_ |  | |  ||  | |  ||   | __ 
# |    ___||    __  ||  |_|  ||  |_|  ||   ||  |
# |   |    |   |  | ||       ||       ||   |_| |
# |___|    |___|  |_||_______||_______||_______|

import numpy as np
from typing import Callable, Union, Any, Tuple
from froog.tensor import Tensor
from froog.utils import mask_like

def jacobian(model: Callable[[Tensor], Tensor], input: Tensor) -> np.ndarray:
  output = model(input)
  ji = input.data.reshape(-1).shape[-1]  # jacobian of input
  jo = output.data.reshape(-1).shape[-1] # jacobian of output
  J = np.zeros((jo, ji), dtype=np.float32)
  for o in range(jo):
    o_scalar = Tensor(mask_like(output.data, o, 1.)).mul(output).sum()
    o_scalar.backward()
    for i, grad in enumerate(input.grad.data.reshape(-1)):
      J[o,i] = grad
  return J

def numerical_jacobian(model: Callable[[Tensor], Tensor], input: Tensor, eps: float = 1e-6) -> np.ndarray:
#     """
#     https://timvieira.github.io/blog/post/2017/04/21/how-to-test-gradient-implementations/
#     Computes :
#         First-order partial derivatives using Finite-Difference Approximation with Central Difference Method (CDM)
#     Params:
#         model : A froog model
#         input : An input
#         eps   : Perturbation step
#     Returns:
#         NJ    : an approx. of the Jacobian
#     """
  output = model(input)
  ji = input.data.reshape(-1).shape[-1]
  jo = output.data.reshape(-1).shape[-1]
  NJ = np.zeros((jo, ji), dtype=np.float32)
  for i in range(ji):
    eps_perturb = mask_like(input.data, i, mask_value = eps)
    output_perturb_add = model(Tensor(input.data + eps_perturb)).data.reshape(-1)
    output_perturb_sub = model(Tensor(input.data - eps_perturb)).data.reshape(-1)
    grad_approx = ((output_perturb_add) - (output_perturb_sub)) / (2*eps) # CDM: (f(x + h) - f(x - h)) / (2 * h)
    NJ[:,i] = grad_approx
  return NJ

def gradcheck(model: Callable[[Tensor], Tensor], input: Tensor, eps: float = 1e-06, atol: float = 1e-5, rtol: float = 0.001) -> bool:
  """
  Checks whether computed gradient is close to numerical approximation of the Jacobian
  Params:
    model       : froog model   
    eps         : eps used to see if gradient is within tolerances
    atol        : absolute tolerance
    rtol        : relative tolerance 
  Returns:
    test_passed : bool of whether the test passed
  """
  NJ = numerical_jacobian(model, input, eps)
  J = jacobian(model, input)
  return np.allclose(J, NJ, atol=atol, rtol=rtol)


import froog.optim
import froog.tensor
import froog.utils

# Import GPU packages
import froog.gpu.cl.cl_utils

# Try to import Metal utils if available
try:
    import froog.gpu.metal.metal_utils
except ImportError:
    pass

# Import device management functions
from froog.gpu import (
    Device, OpenCLDevice, get_device, set_device,
    upload_tensor, download_tensor, is_buffer,
    allocate_buffer, synchronize, get_available_devices
)

# Try to import Metal device if available
try:
    from froog.gpu.metal import MetalDevice
    __all__ = [
        'Device', 'OpenCLDevice', 'MetalDevice', 
        'get_device', 'set_device', 'upload_tensor', 
        'download_tensor', 'is_buffer', 
        'allocate_buffer', 'synchronize', 'get_available_devices'
    ]
except ImportError:
    __all__ = [
        'Device', 'OpenCLDevice', 
        'get_device', 'set_device', 'upload_tensor', 
        'download_tensor', 'is_buffer', 
        'allocate_buffer', 'synchronize', 'get_available_devices'
    ]


#  _______  ______    _______  _______  _______ 
# |       ||    _ |  |       ||       ||       |
# |    ___||   | ||  |   _   ||   _   ||    ___|
# |   |___ |   |_||_ |  | |  ||  | |  ||   | __ 
# |    ___||    __  ||  |_|  ||  |_|  ||   ||  |
# |   |    |   |  | ||       ||       ||   |_| |
# |___|    |___|  |_||_______||_______||_______|

import numpy as np
from typing import Tuple, Union, Optional, Any, Callable
from froog.tensor import Function, register, Tensor
from froog.utils import im2col, col2im

# *****************************************************
#     ____  ___   _____ __________   ____  ____  _____
#    / __ )/   | / ___//  _/ ____/  / __ \/ __ \/ ___/
#   / __  / /| | \__ \ / // /      / / / / /_/ /\__ \ 
#  / /_/ / ___ |___/ // // /___   / /_/ / ____/___/ / 
# /_____/_/  |_/____/___/\____/   \____/_/    /____/  
#
# **************** Basic Operations ***************
# - grad_output is the gradient of the loss with respect to the output of the operation.

class Add(Function):# x.add(y)
  @staticmethod     # @staticmethod doesn't require an instance of Add to work, so you can do x.add(y)
  def forward(ctx: Any, x: np.ndarray, y: np.ndarray) -> np.ndarray:
    # Check if we have GPU buffers
    is_metal_buffer = lambda x: hasattr(x, '__pyobjc_object__') or str(type(x)).find('Buffer') >= 0
    if is_metal_buffer(x) or is_metal_buffer(y):
      # Import get_buffer_data helper for Metal buffers
      try:
        from froog.gpu.buffer_utils import get_buffer_data
        x_data = get_buffer_data(x)
        y_data = get_buffer_data(y)
        ctx.save_for_backward(x_data, y_data)
        return x_data + y_data
      except ImportError:
        print("Warning: buffer_utils not available")
        # Fall back to regular implementation
        ctx.save_for_backward(x, y)
        return x + y
    
    # Regular implementation
    ctx.save_for_backward(x, y)
    return x + y
  
  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    return grad_output, grad_output 
register("add", Add)

class Sub(Function): # x.sub(y)
  @staticmethod
  def forward(ctx: Any, x: np.ndarray, y: np.ndarray) -> np.ndarray:
    return x-y

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    return grad_output, -grad_output
register('sub', Sub)

class Mul(Function): # x.mul(y)
  @staticmethod
  def forward(ctx: Any, x: np.ndarray, y: np.ndarray) -> np.ndarray:
    # Check if we have GPU buffers
    is_metal_buffer = lambda x: hasattr(x, '__pyobjc_object__') or str(type(x)).find('Buffer') >= 0
    if is_metal_buffer(x) or is_metal_buffer(y):
      # Import get_buffer_data helper for Metal buffers
      try:
        from froog.gpu.buffer_utils import get_buffer_data, buffer_mul
        x_data = get_buffer_data(x)
        y_data = get_buffer_data(y)
        ctx.save_for_backward(x_data, y_data)
        return buffer_mul(x, y)
      except Exception as e:
        print(f"Error in Mul.forward with buffer: {e}")
        # Fall back to CPU implementation if buffer handling fails
        from froog.gpu import get_device
        device = get_device()
        if device:
          x_cpu = device.download_tensor(x) if is_metal_buffer(x) else x
          y_cpu = device.download_tensor(y) if is_metal_buffer(y) else y
          ctx.save_for_backward(x_cpu, y_cpu)
          result = x_cpu * y_cpu
          return device.upload_tensor(result)
        raise
    
    # Standard CPU implementation
    ctx.save_for_backward(x, y)
    return x * y

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    x, y = ctx.saved_tensors
    return y * grad_output, x * grad_output
register("mul", Mul)

class Sum(Function): # x.sum()
  """
  reduce op
  reduces its input tensor to a single value by summing all the elements
  """
  @staticmethod
  def forward(ctx: Any, input: np.ndarray) -> np.ndarray:
    # Check if we have a GPU buffer
    is_metal_buffer = lambda x: hasattr(x, '__pyobjc_object__') or str(type(x)).find('Buffer') >= 0
    if is_metal_buffer(input):
      # Use buffer utilities
      try:
        from froog.gpu.buffer_utils import get_buffer_data, buffer_sum
        input_data = get_buffer_data(input)
        ctx.save_for_backward(input_data)
        ctx.input_shape = input_data.shape
        return buffer_sum(input)
      except Exception as e:
        print(f"Error in Sum.forward with buffer: {e}")
        # Fall back to CPU implementation
        from froog.gpu import get_device
        device = get_device()
        if device:
          input_cpu = device.download_tensor(input)
          ctx.save_for_backward(input_cpu)
          ctx.input_shape = input_cpu.shape
          result = np.array([np.sum(input_cpu)])
          return device.upload_tensor(result)
        raise
    
    # Standard CPU implementation
    ctx.save_for_backward(input)
    ctx.input_shape = input.shape
    return np.array([np.sum(input)])

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    (input,) = ctx.saved_tensors
    return grad_output * np.ones_like(input)
register("sum", Sum)

class Pow(Function): # x.pow(y)
  @staticmethod
  def forward(ctx: Any, x: np.ndarray, y: np.ndarray) -> np.ndarray:
    ctx.save_for_backward(x, y)
    return x ** y
  
  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    x, y = ctx.saved_tensors
    return y * (x**(y-1.0)) * grad_output, (x**y) * np.log(x) * grad_output # power rule, d/dx (y^x)
register("pow", Pow)


# *******************************               
#    ______________  _____  ___
#   / ____/ ____/  |/  /  |/  /
#  / / __/ __/ / /|_/ / /|_/ / 
# / /_/ / /___/ /  / / /  / /  
# \____/_____/_/  /_/_/  /_/   
# 
# ******* GEMM ops *******          
                             
class Dot(Function):  # x.dot(y)
  @staticmethod
  def forward(ctx: Any, input: np.ndarray, weight: np.ndarray) -> np.ndarray:
    ctx.save_for_backward(input, weight)
    
    # Check if we're working with GPU buffers
    try:
        from froog.tensor import is_buffer
        from froog.gpu import download_tensor
        
        # Convert any GPU buffers to CPU for the operation
        if is_buffer(input):
            input_cpu = download_tensor(input)
        else:
            input_cpu = input
            
        if is_buffer(weight):
            weight_cpu = download_tensor(weight)
        else:
            weight_cpu = weight
            
        return input_cpu.dot(weight_cpu)
    except Exception as e:
        import traceback
        print(f"Error in dot operation: {str(e)}")
        print(f"  Self: {input}")
        print(f"  Arg 0: {weight}")
        print(f"  Kwargs: {{}}")
        traceback.print_exc()
        # Try the original method as fallback
        return input.dot(weight)

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    input, weight = ctx.saved_tensors
    
    # Convert GPU buffers to CPU if needed
    try:
        from froog.tensor import is_buffer
        from froog.gpu import download_tensor
        
        if is_buffer(input):
            input_cpu = download_tensor(input)
        else:
            input_cpu = input
            
        if is_buffer(weight):
            weight_cpu = download_tensor(weight)
        else:
            weight_cpu = weight
            
        if is_buffer(grad_output):
            grad_output_cpu = download_tensor(grad_output)
        else:
            grad_output_cpu = grad_output
            
        return grad_output_cpu.dot(weight_cpu.T), input_cpu.T.dot(grad_output_cpu)
    except Exception as e:
        print(f"Error in dot backward: {str(e)}")
        # Fallback
        return grad_output.dot(weight.T), input.T.dot(grad_output)
register('dot', Dot)
register('matmul', Dot)


# ***********************************************************
#    _____ ______  _______  __    ______   ____  ____  _____
#   / ___//  _/  |/  / __ \/ /   / ____/  / __ \/ __ \/ ___/
#   \__ \ / // /|_/ / /_/ / /   / __/    / / / / /_/ /\__ \ 
#  ___/ // // /  / / ____/ /___/ /___   / /_/ / ____/___/ / 
# /____/___/_/  /_/_/   /_____/_____/   \____/_/    /____/  
#
# ************************ nn ops ***********************              
                                                          
class ReLU(Function): 
  @staticmethod
  def forward(ctx: Any, input: np.ndarray) -> np.ndarray:
    ctx.save_for_backward(input)
    return np.maximum(input, 0)                     # relu(x) = max(0,x)

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    input, = ctx.saved_tensors
    grad_input = grad_output * (input >= 0)
    return grad_input
register("relu", ReLU)

class Sigmoid(Function): 
  @staticmethod
  def forward(ctx: Any, input: np.ndarray) -> np.ndarray:
    ctx.save_for_backward(input)
    ret = 1/(1 + np.exp(-input))                    # sigmoid(x) = 1 / (1 + exp(-x))
    return ret 

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    ret, = ctx.saved_tensors
    grad_input = grad_output * (ret * (1 - ret))    # just take the derivative of sigmoid
    return grad_input
register("sigmoid", Sigmoid)

class DropoutLayer:
  """
  Dropout layer that randomly sets a fraction of input units to 0 during training time.
  pytorch version: https://pytorch.org/docs/stable/generated/torch.nn.Dropout.html
  """
  def __init__(self, p: float = 0.5) -> None:
    self.p = p
    self.training = True

  def __call__(self, x):
    # build a CPU‐side random mask of the same shape as the tensor x
    mask_np = (np.random.rand(*x.shape) >= self.p).astype(np.float32) / (1.0 - self.p)
    from froog.tensor import Tensor
    mask_t = Tensor(mask_np)
    if getattr(x, "is_gpu", False): mask_t = mask_t.to_gpu()
    return x.mul(mask_t)

class Dropout(Function):
  @staticmethod
  def forward(ctx: Any, input: np.ndarray, p: float = 0.5, training: bool = True) -> np.ndarray:
    if not training: return input
    # create a binary mask with probability (1-p) of being 1
    # scale by 1/(1-p) to keep expectation same
    ctx.training = training
    mask = (np.random.rand(*input.shape) >= p).astype(np.float32) / (1.0 - p if p < 1.0 else 1e-9) # avoid division by zero if p is 1.0
    ctx.save_for_backward(mask)
    return input * mask

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    if not ctx.training: return grad_output
    mask, = ctx.saved_tensors
    return grad_output * mask
register("dropout", Dropout)

class Reshape(Function):
  @staticmethod
  def forward(ctx: Any, x: np.ndarray, shape: Tuple[int, ...]) -> np.ndarray:
    ctx.save_for_backward(x.shape)
    return x.reshape(shape)

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    in_shape, = ctx.saved_tensors
    return grad_output.reshape(in_shape)
register('reshape', Reshape)

class Pad2D(Function):
  """
  The first element (0,0) corresponds to padding along the batch dimension, which indicates no padding on both sides (0 elements added).
  """
  @staticmethod
  def forward(ctx: Any, x: np.ndarray, padding: Optional[Tuple[int, int, int, int]] = None) -> np.ndarray: 
    if padding is None:
      padding = (0, 0, 0, 0)
    return np.pad(x, ((0,0), (0,0), (padding[0], padding[1]), (padding[2], padding[3]))) # (top, bottom, left, right)

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    raise Exception("write this")
register('pad2d', Pad2D)

class LogSoftmax(Function):
  """
  converts a vector of numbers into a vector of probabilities
  probabilities of each value are proportional to the scale of each value 
  """
  @staticmethod
  def forward(ctx: Any, input: np.ndarray) -> np.ndarray:
    def logsumexp(x: np.ndarray) -> np.ndarray:
      c = x.max(axis=1)
      return c + np.log(np.exp(x - c.reshape((-1, 1))).sum(axis=1)) # axis=1 refers to the columns

    output = input - logsumexp(input).reshape((-1, 1))
    ctx.save_for_backward(output)
    return output

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    (output,) = ctx.saved_tensors
    return grad_output - np.exp(output)*(grad_output.sum(axis=1).reshape((-1, 1)))
register("logsoftmax", LogSoftmax)


# *************************************************
#    __________  _   ___    __   ____  ____  _____
#   / ____/ __ \/ | / / |  / /  / __ \/ __ \/ ___/
#  / /   / / / /  |/ /| | / /  / / / / /_/ /\__ \ 
# / /___/ /_/ / /|  / | |/ /  / /_/ / ____/___/ / 
# \____/\____/_/ |_/  |___/   \____/_/    /____/  
#
# ****************** conv ops *****************

class Conv2D(Function): # TODO: understand group splits
  @staticmethod
  def forward(ctx: Any, x: np.ndarray, w: np.ndarray, stride: Union[int, Tuple[int, int]] = 1, groups: int = 1) -> np.ndarray:
    """
    https://github.com/vdumoulin/conv_arithmetic/blob/master/README.md
    WARNING: doesn't handle padding or strides yet
    Args:
      x.shape[0]                                                                        --> number of input examples (batch size)
      cout                                                                                        --> number of output channels
      x.shape[2]-(H-1)                                                  --> non-padded height of conv output, need to subtract because this is an unpadded conv
      x.shape[3]-(W-1)                                                  --> width of output
    Shape: 
      (a, b, c, d)(e, f, g, h)      --> (a, e, c-(g-1), d-(h-1)) 
      in general, output x and y = [(W−K+2P)/S]+1
    """
    ctx.stride = stride
    ctx.groups = groups
    
    if isinstance(ctx.stride, int):                                                                       # ctx stores function params
      ctx.stride = (ctx.stride, ctx.stride)

    # Check if we're working with GPU buffers and convert to CPU
    try:
        from froog.tensor import is_buffer
        from froog.gpu import download_tensor
        
        # Convert input to CPU if it's a GPU buffer
        if is_buffer(x):
            x_cpu = download_tensor(x)
        else:
            x_cpu = x
            
        # Convert weight to CPU if it's a GPU buffer
        if is_buffer(w):
            w_cpu = download_tensor(w)
        else:
            w_cpu = w
            
        # Now use the CPU tensors for the rest of the computation
        cout, cin, H, W = w_cpu.shape
        
        tw = w_cpu.reshape(cout, -1).T                                                                        # slice of kernel 
        y_stride, x_stride = ctx.stride

        bs,cin_,oy,ox = x_cpu.shape[0], x_cpu.shape[1], (x_cpu.shape[2]-(H-y_stride))//y_stride, (x_cpu.shape[3]-(W-x_stride))//x_stride
        assert cin*ctx.groups == cin_                                                                         # ensures that the channel dimensions match appropriately for grouping
        assert cout % ctx.groups == 0                                                                         # ensures that the number of output channels can be evenly divided among the groups
        g_w_chans = cout//ctx.groups                                                                          # number of output channels per group

        ctx.save_for_backward(x_cpu, w_cpu)
        ret = np.zeros((bs, cout, oy, ox), dtype=w_cpu.dtype)
        
        for g in range(ctx.groups):
          tw = w_cpu[g*g_w_chans:(g*g_w_chans+g_w_chans)].reshape(g_w_chans, -1).T                           # transformed kernel weights
          for Y in range(oy):
            for X in range(ox):
              iY,iX = Y*y_stride, X*x_stride
              tx = x_cpu[:, g*cin:(g*cin+cin), iY:iY+H, iX:iX+W].reshape(bs, -1)
              ret[:, g*g_w_chans:(g*g_w_chans+g_w_chans), Y, X] += tx.dot(tw)
        return ret
    except Exception as e:
        import traceback
        print(f"Error in conv2d operation: {str(e)}")
        print(f"  Self: {x}")
        print(f"  Arg 0: {w}")
        print(f"  Kwargs: {{stride: {stride}, groups: {groups}}}")
        traceback.print_exc()
        raise

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    try:
        from froog.tensor import is_buffer
        from froog.gpu import download_tensor
        
        # Convert grad_output to CPU if it's a GPU buffer
        if is_buffer(grad_output):
            grad_output_cpu = download_tensor(grad_output)
        else:
            grad_output_cpu = grad_output
            
        x, w = ctx.saved_tensors
        cout, cin, H, W = w.shape
        dx, dw = np.zeros_like(x), np.zeros_like(w)                                         
        y_stride, x_stride = ctx.stride
        g_w_chans = cout//ctx.groups

        for g in range(ctx.groups):
          tw = w[g*g_w_chans:(g*g_w_chans+g_w_chans)].reshape(g_w_chans, -1)
          for Y in range(grad_output_cpu.shape[2]):
            for X in range(grad_output_cpu.shape[3]):
              iY,iX = Y*y_stride, X*x_stride
              gg = grad_output_cpu[:, g*g_w_chans:(g*g_w_chans+g_w_chans), Y, X]                             # current multiply element in chain rule
              tx = x[:, g*cin:(g*cin+cin), iY:iY+H, iX:iX+W].reshape(x.shape[0], -1)                         # slice of tensor at current conv op        
              dw[g*g_w_chans:(g*g_w_chans+g_w_chans)] += gg.T.dot(tx).reshape((g_w_chans,cin,H,W))           # gradient with respect to input
              dx[:, g*cin:(g*cin+cin), iY:iY+H, iX:iX+W] += gg.dot(tw).reshape(dx.shape[0], cin, H, W)       # accumulate gradient with respect to weights 
        return dx, dw
    except Exception as e:
        import traceback
        print(f"Error in conv2d backward: {str(e)}")
        print(f"  Grad Output: {grad_output}")
        traceback.print_exc()
        # Fallback to original implementation
        x, w = ctx.saved_tensors
        cout, cin, H, W = w.shape
        dx, dw = np.zeros_like(x), np.zeros_like(w)                                         
        y_stride, x_stride = ctx.stride
        g_w_chans = cout//ctx.groups

        for g in range(ctx.groups):
          tw = w[g*g_w_chans:(g*g_w_chans+g_w_chans)].reshape(g_w_chans, -1)
          for Y in range(grad_output.shape[2]):
            for X in range(grad_output.shape[3]):
              iY,iX = Y*y_stride, X*x_stride
              gg = grad_output[:, g*g_w_chans:(g*g_w_chans+g_w_chans), Y, X]                                  # current multiply element in chain rule
              tx = x[:, g*cin:(g*cin+cin), iY:iY+H, iX:iX+W].reshape(x.shape[0], -1)                          # slice of tensor at current conv op        
              dw[g*g_w_chans:(g*g_w_chans+g_w_chans)] += gg.T.dot(tx).reshape((g_w_chans,cin,H,W))            # gradient with respect to input
              dx[:, g*cin:(g*cin+cin), iY:iY+H, iX:iX+W] += gg.dot(tw).reshape(dx.shape[0], cin, H, W)        # accumulate gradient with respect to weights 
        return dx, dw
register('conv2d', Conv2D)


class im2ColConv(Function):
  """
  uses im2col
  https://leonardoaraujosantos.gitbook.io/artificial-inteligence/machine_learning/deep_learning/convolution_layer/making_faster
  """

  @staticmethod
  def forward(ctx: Any, x: np.ndarray, w: np.ndarray) -> np.ndarray:
    cout, cin, k_h, k_x = w.shape
    bs, oy, ox = x.shape[0], x.shape[2]-(k_h-1), x.shape[3]-(k_x-1)
    tw = w.reshape(cout, -1).T                                             # each filter flattened into a row
    tx = im2col(x, k_h, k_x)                                               # im2col, turn input into column
    ctx.save_for_backward(tx, w)                                           # save the im2col output
    ret = tx.dot(tw).reshape(bs, oy, ox, cout)                             # now the conv has been transoforned into a GEMM
    return np.moveaxis(ret, [0,1,2,3], [0,2,3,1])                          # reorders the axes (batch size, number of channels, height, width)

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    bs,_,oy,ox = grad_output.shape
    tx, w = ctx.saved_tensors                                              # transformed input, filter weights 
    cout,cin,H,W = w.shape
    tw = w.reshape(cout, -1)                                               # flatten filter and stack onto other channel filters
    gg = np.moveaxis(grad_output, [0,1,2,3], [1,0,2,3]).reshape(cout, -1)  # order correctly
    dw = gg.dot(tx).reshape(w.shape)                                       # compute gradient of weight
    dxi = gg.T.dot(tw)                                                     # compute gradient of input
    dx = col2im(dxi, H, W, oy+(H-1), ox+(W-1))                             # turn columns back into image shape
    return dx, dw
register('im2col2dconv', im2ColConv)


# *************************************************
#     ____  ____  ____  __       ____  ____  _____
#    / __ \/ __ \/ __ \/ /      / __ \/ __ \/ ___/
#   / /_/ / / / / / / / /      / / / / /_/ /\__ \ 
#  / ____/ /_/ / /_/ / /___   / /_/ / ____/___/ / 
# /_/    \____/\____/_____/   \____/_/    /____/  
#
# **************** pooling ops ***************

def stack_for_pool(x: np.ndarray, pool_y: int, pool_x: int) -> np.ndarray:
  my, mx = (x.shape[2]//pool_y)*pool_y, (x.shape[3]//pool_x)*pool_x        # ensures input tensor can be evenly divided into 2x2 blocks for max pooling
  stack = []
  cropped_x = x[:, :, :my, :mx]                                            # crop input so 2x2 max pool can be taken
  for Y in range(pool_y):
      for X in range(pool_x):
        stack.append(cropped_x[:, :, Y::pool_y, X::pool_x][None])          # ::2 so 2x2 goes to next pool, [None] is numpy way to add an extra dimension so we can concatenate
  return np.concatenate(stack, axis=0)                                     # put all into one row


def unstack_for_pool(fxn: Callable[[int], np.ndarray], s: Tuple[int, ...], py: int, px: int) -> np.ndarray:
  max_y, max_x = (s[2]//py)*py, (s[3]//px)*px                              # get shape that allows (pool_size_y,pool_size_x) max pool
  ret = None
  for Y in range(py):
    for X in range(px):
      level_w_new_grad = fxn(Y*px+X)
      if X == 0 and Y == 0:                                                # pool of zero size
        ret = np.zeros(s, dtype=level_w_new_grad.dtype)
      if ret is not None:
        ret[:, :, Y:max_y:py, X:max_x:px] = level_w_new_grad
  return ret if ret is not None else np.zeros(s)


class MaxPool2D(Function):
  @staticmethod
  def forward(ctx: Any, x: np.ndarray, kernel_size: Tuple[int, int] = (2,2)) -> np.ndarray:
    ctx.kernel_size = kernel_size
    stack = stack_for_pool(x, *kernel_size)
    idx_of_max = np.argmax(stack, axis=0)
    ctx.save_for_backward(idx_of_max, x.shape)
    return np.max(stack, axis=0)

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    """
    Distributes the gradient from the output of the max pooling layer to its inputs
    The purpose of (idxs == idx) is to generate a boolean mask indicating the locations of the maximum values in each 2x2 block of the original input
    The expression (Y*2+X) is a way to iterate through the four possible positions within the kernel block: e.g. (0,0), (0,1), (1,0), and (1,1), which get mapped to the indices 0, 1, 2, and 3 
    """
    idxs, s = ctx.saved_tensors
    return unstack_for_pool(lambda idx: grad_output * (idxs == idx),  s, *ctx.kernel_size)
register('max_pool2d', MaxPool2D)

class AvgPool2D(Function):
  @staticmethod
  def forward(ctx: Any, x: np.ndarray, kernel_size: Tuple[int, int] = (2,2)) -> np.ndarray:
    ctx.kernel_size = kernel_size
    stack = stack_for_pool(x, *kernel_size)
    ctx.save_for_backward(x.shape)
    return np.mean(stack, axis=0)

  @staticmethod
  def backward(ctx: Any, grad_output: np.ndarray) -> np.ndarray:
    s, = ctx.saved_tensors
    py, px = ctx.kernel_size                                  # kernel_size passed from forward context
    my, mx = (s[2]//py)*py, (s[3]//px)*px
    ret = np.zeros(s, dtype=grad_output.dtype)
    for Y in range(py):
      for X in range(px):
        ret[:, :, Y:my:py, X:mx:px] = grad_output / py / px   # divide by avg of pool, e.g. for 2x2 pool /= 4
    return ret
register('avg_pool2d', AvgPool2D)

# *************************************
#     _   ___   __   ____  ____  _____
#    / | / / | / /  / __ \/ __ \/ ___/
#   /  |/ /  |/ /  / / / / /_/ /\__ \ 
#  / /|  / /|  /  / /_/ / ____/___/ / 
# /_/ |_/_/ |_/   \____/_/    /____/  
#
# ************* nn ops ************   

def Linear(*x: int) -> np.ndarray:
  # random Glorot initialization
  ret = np.random.uniform(-1., 1., size=x)/np.sqrt(np.prod(x))
  return ret.astype(np.float32)

def swish(x: Tensor) -> Tensor:
  return x.mul(x.sigmoid())

class BatchNorm2D:
  """
  __call__ follows the formula from the link below
  pytorch version: https://pytorch.org/docs/stable/generated/torch.nn.BatchNorm2d.html

  self.weight       = γ
  self.bias         = β
  self.running_mean = E[x] 
  self.running_var  = Var[x]

  the reshaping step ensures that each channel of the input has its 
  own separate set of parameters (mean, variance, weight, and bias)

  self.running_mean has shape [num_channels].
  self.running_mean.reshape(shape=[1, -1, 1, 1]) reshapes it to [1, num_channels, 1, 1]
  """
  def __init__(self, sz: int, eps: float = 0.001) -> None:
    self.eps = eps
    self.weight = Tensor.zeros(sz)
    self.bias = Tensor.zeros(sz)

    # TODO: need running_mean and running_var
    self.running_mean = Tensor.zeros(sz)
    self.running_var = Tensor.zeros(sz)
    self.num_batches_tracked = Tensor.zeros(1)

  def __call__(self, x: Tensor) -> Tensor:
    x = x.sub(self.running_mean.reshape(shape=[1, -1, 1, 1]))
    x = x.mul(self.weight.reshape(shape=[1, -1, 1, 1]))
    x = x.div(self.running_var.add(Tensor([self.eps], gpu=x.gpu)).reshape(shape=[1, -1, 1, 1]).sqrt())
    x = x.add(self.bias.reshape(shape=[1, -1, 1, 1]))
    return x


#!/usr/bin/env python3
# this file specifies how the froog package is installed, including any necessary dependencies required to run

import os
from setuptools import setup

directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(directory, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

setup(name='froog',
      version='0.5.1',
      description='tensor library with opencl and metal support',
      author='Kevin Buhler',
      license='MIT',
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages = ['froog'],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
      ],
      install_requires=['numpy', 'requests', 'matplotlib'],
      python_requires='>=3.8',
      include_package_data=True)